﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using XrMapLib;
using AxXrMapLib;

namespace MerageSHPFiles
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        private XrShapeTypeEnum GetShapeType()
        {
            String FileName = lbSHPFiles.Items[0].ToString();
            Xr.Layers.AddShapeMapLayer(LAYERNAME, FileName);
            Xr.WaitForAllConnections();
            
            ShapeMapLayer Lyr = Xr.Layers.GetLayerAsShapeMap(LAYERNAME);
            XrShapeTypeEnum Result = Lyr.ShapeType;
            Xr.Layers.RemoveAllLayer();
            
            return Result;
        }

        private FieldSetDefine BuildFieldDefine(CheckedListBox clb)
        {
            if (lbSHPFiles.Items.Count == 0) return null;

            String FileName = lbSHPFiles.Items[0].ToString();
            if (Xr.Layers.AddShapeMapLayer(LAYERNAME, FileName))
            {
                Xr.WaitForAllConnections();

                ShapeMapLayer Lyr = Xr.Layers.GetLayerAsShapeMap(LAYERNAME);
                if (Lyr == null) return null;

                IAttributeTable AT = Lyr.AttributeTable;
                FieldSet FS = AT.FieldSet;
                FieldSetDefine FSD = new XrMapLib.FieldSetDefine();

                int CountFields = FS.FieldCount;
                if (CountFields > 0)
                {
                    CheckedListBox.CheckedItemCollection CheckedItems = clb.CheckedItems;

                    for (int iField = 0; iField < CountFields; ++iField)
                    {
                        String FieldName = FS.GetFieldName(iField);
                        if (CheckedItems.Contains(FieldName))
                        {
                            XrFieldTypeEnum FieldType = FS.GetFieldType(iField);
                            FieldSize fieldSize = fieldSize_[FieldName];
                            int FieldWidth = fieldSize.Width;
                            int FieldDecimal = fieldSize.Decimal;

                            if (FieldType == XrFieldTypeEnum.XrDouble || FieldType == XrFieldTypeEnum.XrFloat)
                            {
                                FSD.AddFloatField(FieldName, FieldWidth, FieldDecimal);
                            }
                            else if (FieldType == XrFieldTypeEnum.XrInteger || FieldType == XrFieldTypeEnum.XrShort || FieldType == XrFieldTypeEnum.XrVeryShort)
                            {
                                FSD.AddIntegerField(FieldName, FieldWidth);
                            }
                            else if (FieldType == XrFieldTypeEnum.XrString)
                            {
                                FSD.AddStringField(FieldName, FieldWidth);
                            }
                        }
                    }

                    Xr.Layers.RemoveAllLayer();

                    return FSD;
                }
            }

            return null;
        }

        private const String LAYERNAME = "lyr";

        public bool FillFieldNames(AxXr Xr, CheckedListBox clb)
        {
            clb.BeginUpdate();
            clb.Items.Clear();

            XrMapLib.ShapeMapLayer ML = Xr.Layers.GetLayerAsShapeMap(LAYERNAME);
            if (ML == null) return false;

            XrMapLib.AttributeTable AT = ML.AttributeTable;

            int cntFields = AT.FieldSet.FieldCount;
            for (int iField = 0; iField < cntFields; ++iField)
            {
                String FieldName = AT.FieldSet.GetFieldName(iField);
                clb.Items.Add(AT.FieldSet.GetFieldName(iField));

                FieldSize fieldSize = new FieldSize();
                fieldSize.Width = AT.FieldSet.GetFieldWidth(iField);
                fieldSize.Decimal = AT.FieldSet.GetFieldDecimal(iField);
                fieldSize_[FieldName] = fieldSize;
            }

            clb.EndUpdate();

            return true;
        }

        private class FieldSize
        {
            public int Width;
            public int Decimal;
        }

        private Dictionary<String, FieldSize> fieldSize_ = new Dictionary<string, FieldSize>();

        public bool CheckFieldNames(AxXrMapLib.AxXr Xr, CheckedListBox clb)
        {
            XrMapLib.ShapeMapLayer ML = Xr.Layers.GetLayerAsShapeMap(LAYERNAME);
            if (ML == null) return false;

            XrMapLib.AttributeTable AT = ML.AttributeTable;

            int cntFields = AT.FieldSet.FieldCount;
            for (int iField = 0; iField < cntFields; ++iField)
            {
                String FieldName = AT.FieldSet.GetFieldName(iField);
                if (!clb.Items.Contains(FieldName))
                {
                    return false;
                }
                else
                {
                    FieldSize fieldSize = fieldSize_[FieldName];

                    int FieldWidth = AT.FieldSet.GetFieldWidth(iField);
                    int FieldDecimal = AT.FieldSet.GetFieldDecimal(iField);

                    fieldSize.Width = Math.Max(fieldSize.Width, FieldWidth);
                    fieldSize.Decimal = Math.Max(fieldSize.Decimal, FieldDecimal);
                }
            }

            return true;
        }

        private void btnAddSHP_Click(object sender, EventArgs e)
        {
            DialogResult DR = openFileDialog.ShowDialog();
            if (DR == DialogResult.OK)
            {
                String FileName = openFileDialog.FileName;

                if (lbSHPFiles.Items.Contains(FileName))
                {
                    MessageBox.Show("이미 추가된 레이어입니다.");
                    return;
                }

                if (lbSHPFiles.Items.Count == 0)
                {
                    clbFieldList.Items.Clear();
                }
                
                lbSHPFiles.Items.Add(FileName);

                if (Xr.Layers.AddShapeMapLayer(LAYERNAME, FileName))
                {
                    Xr.WaitForAllConnections();

                    if (clbFieldList.Items.Count == 0)
                    {
                        FillFieldNames(Xr, clbFieldList);
                    }
                    else
                    {
                        if (!CheckFieldNames(Xr, clbFieldList))
                        {
                            MessageBox.Show("필드구조가 다르므로 병합할 수 없는 레이어입니다.");
                            lbSHPFiles.Items.RemoveAt(lbSHPFiles.Items.Count - 1);
                        }
                    }

                    Xr.Layers.RemoveAllLayer();
                }

                lblCount.Text = lbSHPFiles.Items.Count + " File(s) Added";
            }
        }

        private void btnDelSHP_Click(object sender, EventArgs e)
        {
            int SelectedIndex = lbSHPFiles.SelectedIndex;
            if (SelectedIndex != -1)
            {
                lbSHPFiles.Items.RemoveAt(SelectedIndex);

                if (lbSHPFiles.Items.Count == 0)
                {
                    clbFieldList.Items.Clear();
                }

                lblCount.Text = lbSHPFiles.Items.Count + " File(s) Added";
            }
        }

        private void btnResultFile_Click(object sender, EventArgs e)
        {
            DialogResult DR = saveFileDialog.ShowDialog();
            if (DR == DialogResult.OK)
            {
                String FileName = saveFileDialog.FileName;
                tbResultFile.Text = FileName;
            }
        }

        private XrMapLib.Coordinate GetShapeCoord(XrMapLib.IShapeRow row)
        {
            XrMapLib.Coordinate result = new XrMapLib.Coordinate();
            XrMapLib.Coordinate coord = row.GetVertex(0);
            result.X = coord.X;
            result.Y = coord.Y;

            return result;
        }

        private XrMapLib.CoordinateListCollection GetShapeCoordList(IShapeRow row)
        {
            XrMapLib.CoordinateListCollection CLC = new XrMapLib.CoordinateListCollection();

            int cntParts = row.PartCount;
            for (int iPart = 0; iPart < cntParts; ++iPart)
            {
                int cntVertices = row.GetVertexCountFromPart(iPart);
                XrMapLib.CoordinateList CL = new XrMapLib.CoordinateList();
                for (int iVertex = 0; iVertex < cntVertices; ++iVertex)
                {
                    XrMapLib.Coordinate coord = row.GetVertexFromPart(iPart, iVertex);
                    CL.Add(coord.X, coord.Y);
                }

                CLC.Add(CL);
            }

            return CLC;
        }

        private XrMapLib.ValueList GetAttributeValueList(XrMapLib.IAttributeRow row, int fieldsCount, CheckedListBox.CheckedIndexCollection cic)
        {
            XrMapLib.ValueList values = new XrMapLib.ValueList();

            for (int iField = 0; iField < fieldsCount; ++iField)
            {
                if(cic.Contains(iField))
                {
                    String value = row.GetValueAsString(iField);
                    values.AddValueByString(value);
                }
            }

            return values;
        }

        private void btnRun_Click(object sender, EventArgs e)
        {
            FieldSetDefine FSD = BuildFieldDefine(clbFieldList);
            if (FSD == null)
            {
                MessageBox.Show("필드 구조를 얻을 수 없습니다.");
                return;
            }

            XrShapeTypeEnum ShapeType = GetShapeType();

            XrMapLib.Tool tool = new XrMapLib.Tool();
            EsriShpFileHandler SFH = tool.CreateEsriShpFileHandler(tbResultFile.Text, ShapeType, FSD);
            if (SFH == null)
            {
                MessageBox.Show("SHP 파일을 생성할 수 없습니다.");
                return;
            }

            CheckedListBox.CheckedIndexCollection CIC = clbFieldList.CheckedIndices;

            int CountLayers = lbSHPFiles.Items.Count;
            for (int iLyr = 0; iLyr < CountLayers; iLyr++)
            {
                String FileName = lbSHPFiles.Items[iLyr].ToString();
                if (Xr.Layers.AddShapeMapLayer(LAYERNAME, FileName))
                {
                    Xr.WaitForAllConnections();
                    ShapeMapLayer SMLyr = Xr.Layers.GetLayerAsShapeMap(LAYERNAME);
                    ShapeTable STbl = SMLyr.ShapeTable;
                    AttributeTable ATbl = SMLyr.AttributeTable;
                    int CountRows = STbl.RowCount;
                    int fieldsCount = ATbl.FieldSet.FieldCount;

                    for (int iFID = 0; iFID < CountRows; iFID++)
                    {
                        XrMapLib.IShapeRow SRow = STbl.GetRow(iFID);
                        XrMapLib.IAttributeRow ARow = ATbl.GetRow(iFID);

                        if (SRow.Load() && ARow.Load())
                        {
                            XrMapLib.ValueList VL = GetAttributeValueList(ARow, fieldsCount, CIC);
                            bool bOK = false;

                            if(ShapeType == XrShapeTypeEnum.XrPoint)
                            {
                                XrMapLib.Coordinate COORD = GetShapeCoord(SRow);
                                bOK = SFH.AddPoint(COORD.X, COORD.Y, VL);
                            }
                            else
                            {
                                XrMapLib.CoordinateListCollection CLC = GetShapeCoordList(SRow);
                                bOK = SFH.AddMultiPart(CLC, VL);
                            }
                            
                            SRow.Unload();
                            ARow.Unload();

                            if (!bOK)
                            {
                                MessageBox.Show("SHP 파일 쓰기 실패");
                                return;
                            }
                        }
                    }

                    Xr.Layers.RemoveAllLayer();
                }
                else
                {
                    MessageBox.Show("레이어 추가 실패: " + FileName);
                    return;
                }
            }

            SFH.Close();
            MessageBox.Show("완료");
            Close();
        }
    }
}
